"""Tushare utils directory."""
import os
from typing import Tuple
from capinvest_core.app.utils import get_user_cache_directory

def get_log_path() -> str:
    """
    Get the path for Tushare log file.

    Returns:
        str: The path to the Tushare log file.
    """
    log_dir = f"{get_user_cache_directory()}/tushare"
    log_path = f"{log_dir}/capinvest_tushare.log"

    os.makedirs(log_dir, exist_ok=True)

    return log_path

def get_cache_path() -> str:
    """
    Get the path for Tushare cache database.

    Returns:
        str: The path to the Tushare cache database.
    """
    db_dir = f"{get_user_cache_directory()}/tushare"
    db_path = f"{db_dir}/equity.db"

    os.makedirs(db_dir, exist_ok=True)

    return db_path


import logging
import pandas as pd
import tushare as ts
from datetime import (
    date as dateType,
    datetime,
)
from typing import Optional, Union
from capinvest_tushare.utils.tools import setup_logger
from capinvest_tushare.utils.helpers import get_api_key
from capinvest_tushare.utils.tools import normalize_symbol

setup_logger()
logger = logging.getLogger(__name__)

EQUITY_HISTORY_SCHEMA = {
    "date": "TEXT PRIMARY KEY",
    "open": "REAL",
    "high": "REAL",
    "low": "REAL",
    "close": "REAL",
    "volume": "REAL",
    "vwap": "REAL",
    "change": "REAL",
    "change_percent": "REAL",
    "amount": "REAL"
}

def get_from_cache(
        ts_code: str,
        start_date: Union[dateType, str],
        end_date: Union[dateType, str],
        api_key : str = "",
        period: str = "daily",
        use_cache: bool = True,
        adjust: str = ""
    ) -> pd.DataFrame:
    """
    Retrieves historical equity data from a cache or downloads it from a remote source.
    
    Parameters:
        symbol (str): Stock symbol to fetch data for.
        start_date (str): Start date for fetching data in 'YYYYMMDD' format.
        end_date (str): End date for fetching data in 'YYYYMMDD' format.
        period (str): Data frequency, e.g., "daily", "weekly", "monthly".
        adjust (str): Adjustment type, e.g., "qfq" for forward split, "hfq" for backward split.

    Returns:
        DataFrame: DataFrame containing historical equity data.
    """
    from capinvest_tushare.utils.table_cache import TableCache

    # Retrieve data from cache first
    symbol_b, symbol_f, market = normalize_symbol(ts_code)
    cache = TableCache(EQUITY_HISTORY_SCHEMA, table_name=f"{market}{symbol_b}", primary_key="date")
    start_dt = datetime.now().date()
    if isinstance(start_date, str):
        start_dt = datetime.strptime(start_date, "%Y-%m-%d")
    elif isinstance(start_date, dateType):
        start_dt = start_date
    else:
        raise ValueError(f"start_date {type(start_date)} must be a string or datetime object")
    
    end_dt = datetime.now().date()
    if isinstance(end_date, str):
        end_dt = datetime.strptime(end_date, "%Y-%m-%d")
    elif isinstance(end_date, dateType):
        end_dt = end_date
    else:
        raise ValueError(f"end_date {type(start_date)} must be a string or datetime object")

    start = start_dt.strftime("%Y%m%d")
    end = end_dt.strftime("%Y%m%d")
    data_from_cache = cache.fetch_date_range(start, end)
    if not data_from_cache.empty:
        logger.info(f"Getting equity {ts_code} historical data from cache...")
        return data_from_cache

    # If not in cache, download data
    # Download data using AKShare
    data_util_today_df = get_one(ts_code, period=period,api_key=api_key, start_date=start_dt, end_date=end_dt)
    cache.write_dataframe(data_util_today_df)
    
    return cache.fetch_date_range(start, end)

def get_one(
        ts_code : str, 
        start_date: dateType,
        end_date: dateType,
        period: str = "daily",
        use_cache: bool = True, 
        api_key : str = ""
        ) -> pd.DataFrame:
    tushare_api_key = get_api_key(api_key)

    pro = ts.pro_api(tushare_api_key)
    symbol_b, symbol, market = normalize_symbol(ts_code)
    df_data = pd.DataFrame()
    if market == 'HK':
        df_data = pro.hk_daily(ts_code=ts_code)
        logger.info(f"Downloaed historical data (HK) {ts_code}: {len(df_data)}.")
    else:
        df_data = pro.daily(ts_code=ts_code)
        logger.info(f"Downloaed historical data {ts_code}: {len(df_data)}.")

    df_data = df_data.rename(columns={'trade_date':'date', 'vol':'volume', "pct_chg":"change_percent"})
    if 'ts_code' in df_data.columns:
        df_data.drop(columns=['ts_code'], inplace=True)
    return df_data


import logging
import os
import pickle
import sqlite3
import time
from datetime import datetime, timedelta
from typing import Optional

from capinvest_tushare.utils.tools import setup_logger

CACHE_TTL = 60*60  # 60 seconds
setup_logger()
logger = logging.getLogger(__name__)

# Constant TTL strategy
def constant_ttl(now: datetime, ttl_seconds: int) -> datetime:
    return now + timedelta(seconds=ttl_seconds)

# Quarter-based expiry (each quarter is 3 months)
def get_next_quarter_start(dt: datetime) -> datetime:
    month = ((dt.month - 1) // 3 + 1) * 3 + 1
    if month > 12:
        return datetime(dt.year + 1, 1, 1)
    return datetime(dt.year, month, 1)

# Year-based expiry
def get_next_year_start(dt: datetime) -> datetime:
    return datetime(dt.year + 1, 1, 1)

def calculate_cache_ttl(ttl_strategy_func, *args, now=None):
    """
    Generic function to calculate cache TTL using a strategy function.
    
    Args:
        ttl_strategy_func: A function that calculates the TTL end time.
        *args: Arguments for the strategy function.
        now: Optional current time (for testing or simulation).
    
    Returns:
        The calculated TTL expiry time.
    """
    now = now or datetime.now()
    return ttl_strategy_func(now, *args)

class BlobCache:
    def __init__(self, table_name: Optional[str] = None, db_path: Optional[str] = None):
        if table_name is None:
            raise ValueError("Table name must be provided")

        self.table_name = table_name
        self.conn = None
        if db_path is None:
            from capinvest_tushare.utils import get_cache_path
            self.db_path = get_cache_path()
        else:
            os.makedirs(db_path, exist_ok=True)
            db_path = f"{db_path}/equity.db"
            self.db_path = db_path
        self._ensure_db_exists()

    def _ensure_db_exists(self):
        """Ensure the SQLite database and table exist."""
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(f'''
                CREATE TABLE IF NOT EXISTS {self.table_name} (
                    key TEXT PRIMARY KEY,
                    timestamp REAL,
                    data BLOB
                )
            ''')
            conn.commit()

    def load_cached_data(self, symbol:str, report_type, use_cache, get_data, api_key : str = "", *args, **kwargs):
        """Load cached data from SQLite cache or generate new data."""
        from capinvest_tushare.utils.tools import normalize_symbol
        symbol_b, symbol_f, market = normalize_symbol(symbol)
        key = f"{market}{symbol_b}{report_type}"
        now = time.time()
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            if use_cache:
                cursor.execute(f'SELECT timestamp, data FROM {self.table_name} WHERE key=?', (key,))
                row = cursor.fetchone()

                if row:
                    timestamp, data_blob = row
                    stored_date = datetime.fromtimestamp(timestamp)
                    if report_type == "annual":
                        expired_date = calculate_cache_ttl(get_next_year_start, now=stored_date)
                        if now < expired_date.timestamp():
                            logger.info("Loading annual data from SQLite cache...")
                            return pickle.loads(data_blob)
                    elif report_type == "quarter":
                        expired_date = calculate_cache_ttl(get_next_quarter_start, now=stored_date)
                        if now < expired_date.timestamp():
                            logger.info("Loading quarter data from SQLite cache...")
                            return pickle.loads(data_blob)
                    else:
                        if now - timestamp < CACHE_TTL:
                            logger.info("Loading data from SQLite cache...")
                            return pickle.loads(data_blob)

            logger.info(f"Generating new {report_type} data...")
            df = get_data(symbol, report_type, api_key=api_key)

            # 序列化 DataFrame
            data_blob = pickle.dumps(df)

            # 更新或插入缓存
            cursor.execute(f'''
                INSERT OR REPLACE INTO {self.table_name} (key, timestamp, data)
                VALUES (?, ?, ?)
            ''', (key, now, data_blob))

            conn.commit()
            return df



import logging
import pandas as pd
import tushare as ts
from typing import Optional, Literal
from capinvest_tushare.utils.tools import setup_logger
from capinvest_tushare.utils.helpers import get_api_key
from capinvest_tushare.utils.tools import normalize_symbol

setup_logger()
logger = logging.getLogger(__name__)

def get_balance_sheet(
        symbol: str, 
        period: Literal["annual", "quarter"] = "annual",
        limit: Optional[int] = 5,
        use_cache: bool = True,
        api_key : Optional[str] = ""
    ) -> pd.DataFrame:
    from capinvest_tushare.utils.blob_cache import BlobCache
    cache = BlobCache(table_name="balance_sheet")
    data = cache.load_cached_data(symbol, "quarter", use_cache, get_tushare_data, api_key=api_key)
    if data is None:
        return pd.DataFrame()
    else:
        if period == "annual":
            # Filter rows where end_date ends with "1231"
            data = data[data['end_date'].astype(str).str.endswith('1231')]

        # Apply limit if specified
        if limit is not None:
            data = data.head(limit)
        
        return processing_data(data)
def get_tushare_data(
        symbol: str,
        period: str = "annual",
        api_key : Optional[str] = ""
    ) -> pd.DataFrame:
    tushare_api_key = get_api_key(api_key)
    pro = ts.pro_api(tushare_api_key)
    _, _, market = normalize_symbol(symbol)
    if market == 'HK':
        balancesheet_df = pro.hk_balancesheet(ts_code=symbol)
    else:
        balancesheet_df = pro.balancesheet(ts_code=symbol)
        balancesheet_df = balancesheet_df.drop_duplicates(subset='end_date', keep='first')

    
    return balancesheet_df

def processing_data(balancesheet_df: pd.DataFrame) -> pd.DataFrame:
    from capinvest_tushare.utils.helpers import get_fiscal_period
    # logger.info("Processing balance sheet data")
    selected_columns = balancesheet_df[['total_assets', 'total_liab']]
    selected_columns = selected_columns.rename(columns={'total_liab':'total_liabilities'})

    # Extract year from end_date and create fiscal_year column
    selected_columns.loc[:, 'fiscal_year'] = pd.to_datetime(balancesheet_df['end_date']).dt.year
    selected_columns.loc[:, 'period_ending'] = pd.to_datetime(balancesheet_df['end_date'])
    selected_columns.loc[:, 'fiscal_period'] = balancesheet_df['end_type'].apply(get_fiscal_period)
    return selected_columns